"""Low-level server communication."""
import time
import logging
import pprint
import uuid

import pycurl

from landscape.lib.fetch import fetch
from landscape.lib import bpickle
from landscape.log import format_delta
from landscape import SERVER_API, VERSION


class HTTPTransport(object):
    """Transport makes a request to exchange message data over HTTP.

    @param url: URL of the remote Landscape server message system.
    @param pubkey: SSH public key used for secure communication.
    """

    def __init__(self, reactor, url, pubkey=None):
        self._reactor = reactor
        self._url = url
        self._pubkey = pubkey

    def get_url(self):
        """Get the URL of the remote message system."""
        return self._url

    def set_url(self, url):
        """Set the URL of the remote message system."""
        self._url = url

    def _curl(self, payload, computer_id, exchange_token, message_api):
        headers = {"X-Message-API": message_api,
                   "User-Agent": "landscape-client/%s" % VERSION,
                   "Content-Type": "application/octet-stream"}
        if computer_id:
            headers["X-Computer-ID"] = computer_id
        if exchange_token:
            headers["X-Exchange-Token"] = str(exchange_token)
        curl = pycurl.Curl()
        return (curl, fetch(self._url, post=True, data=payload,
                            headers=headers, cainfo=self._pubkey, curl=curl))

    def exchange(self, payload, computer_id=None, exchange_token=None,
                 message_api=SERVER_API):
        """Exchange message data with the server.

        @param payload: The object to send, it must be L{bpickle}-compatible.
        @param computer_id: The computer ID to send the message as (see
            also L{Identity}).
        @param exchange_token: The token that the server has given us at the
            last exchange. It's used to prove that we are still the same
            client.

        @type: C{dict}
        @return: The server's response to sent message or C{None} in case
            of error.

        @note: This code is thread safe (HOPEFULLY).

        """
        spayload = bpickle.dumps(payload)
        start_time = time.time()
        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            logging.debug("Sending payload:\n%s", pprint.pformat(payload))
        try:
            curly, data = self._curl(spayload, computer_id, exchange_token,
                                     message_api)
        except:
            logging.exception("Error contacting the server at %s." % self._url)
            raise
        else:
            logging.info("Sent %d bytes and received %d bytes in %s.",
                         len(spayload), len(data),
                         format_delta(time.time() - start_time))

        try:
            response = bpickle.loads(data)
        except:
            logging.exception("Server returned invalid data: %r" % data)
            return None
        else:
            if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
                logging.debug(
                    "Received payload:\n%s", pprint.pformat(response))

        return response


class FakeTransport(object):
    """Fake transport for testing purposes."""

    def __init__(self, reactor=None, url=None, pubkey=None):
        self._pubkey = pubkey
        self.payloads = []
        self.responses = []
        self._current_response = 0
        self.next_expected_sequence = 0
        self.computer_id = None
        self.exchange_token = None
        self.message_api = None
        self.extra = {}
        self._url = url
        self._reactor = reactor

    def get_url(self):
        return self._url

    def set_url(self, url):
        self._url = url

    def exchange(self, payload, computer_id=None, exchange_token=None,
                 message_api=SERVER_API):
        self.payloads.append(payload)
        self.computer_id = computer_id
        self.exchange_token = exchange_token
        self.message_api = message_api
        self.next_expected_sequence += len(payload.get("messages", []))

        if self._current_response < len(self.responses):
            response = self.responses[self._current_response]
            self._current_response += 1
        else:
            response = []

        result = {"next-expected-sequence": self.next_expected_sequence,
                  "next-exchange-token": unicode(uuid.uuid4()),
                  "messages": response}
        result.update(self.extra)
        return result



import logging


def log_failure(failure, msg=None, logger=None):
    """Log a L{twisted.python.failure.Failure} to the Python L{logging} module.

    The failure should be formatted as a regular exception, but a traceback may
    not be available.

    If C{msg} is passed, it will included before the traceback.
    """
    if logger is None:
        logger = logging
    logger.error(msg, exc_info=(failure.type, failure.value, failure.tb))





class SequenceError(Exception):
    """Raised when the sequence isn't proper for translation to ranges."""


class SequenceRanges(object):
    """High level interface to ranges.

    A ranges list represent a sequence of ordered and non-repeating
    elements into a more compact format, by representing 3 or more
    consecutive entries by a range.

    This means that a sequence such as

        [1, 2, 4, 5, 6, 8, 10, 11, 12, 14]

    becomes

        [1, 2, (4, 6), 8, (10, 12), 14]
    """

    def __init__(self):
        self._ranges = []

    @classmethod
    def from_sequence(cls, sequence):
        obj = cls()
        obj._ranges[:] = sequence_to_ranges(sequence)
        return obj

    @classmethod
    def from_ranges(cls, ranges):
        obj = cls()
        obj._ranges[:] = ranges
        return obj

    def to_sequence(self):
        return list(ranges_to_sequence(self._ranges))

    def to_ranges(self):
        return list(self._ranges)

    def __iter__(self):
        return ranges_to_sequence(self._ranges)

    def __contains__(self, item):
        index = find_ranges_index(self._ranges, item)
        if index < len(self._ranges):
            test = self._ranges[index]
            if isinstance(test, tuple):
                return (test[0] <= item <= test[1])
            return (test == item)
        return False

    def add(self, item):
        add_to_ranges(self._ranges, item)

    def remove(self, item):
        remove_from_ranges(self._ranges, item)


def sequence_to_ranges(sequence):
    """Iterate over range items that compose the given sequence."""

    iterator = iter(sequence)
    try:
        range_start = range_stop = iterator.next()
    except StopIteration:
        return
    while range_start is not None:
        try:
            item = iterator.next()
        except StopIteration:
            item = None
        if item == range_stop + 1:
            range_stop += 1
        else:
            if item is not None and item <= range_stop:
                if item < range_stop:
                    raise SequenceError("Sequence is unordered (%r < %r)" %
                                        (item, range_stop))
                else:
                    raise SequenceError("Found duplicated item (%r)" % (item,))
            if range_stop == range_start:
                yield range_start
            elif range_stop == range_start + 1:
                yield range_start
                yield range_stop
            else:
                yield (range_start, range_stop)
            range_start = range_stop = item


def ranges_to_sequence(ranges):
    """Iterate over individual items represented in a ranges list."""
    for item in ranges:
        if isinstance(item, tuple):
            start, end = item
            if start > end:
                raise ValueError("Range error %d > %d", start, end)
            for item in xrange(start, end + 1):
                yield item
        else:
            yield item


def find_ranges_index(ranges, item):
    """Find the index where an entry *may* be."""
    lo = 0
    hi = len(ranges)
    while lo < hi:
        mid = (lo + hi) // 2
        test = ranges[mid]
        try:
            test = test[1]
        except TypeError:
            pass
        if item > test:
            lo = mid + 1
        else:
            hi = mid
    return lo


def add_to_ranges(ranges, item):
    """Insert item in ranges, reorganizing as needed."""

    index_start = index_stop = index = find_ranges_index(ranges, item)
    range_start = range_stop = item

    ranges_len = len(ranges)

    # Look for duplicates.
    if index < ranges_len:
        test = ranges[index]
        if isinstance(test, tuple):
            if test[0] <= item <= test[1]:
                return
        elif test == item:
            return

    # Merge to the left side.
    while index_start > 0:
        test = ranges[index_start - 1]
        if isinstance(test, tuple):
            if test[1] != range_start - 1:
                break
            range_start = test[0]
        else:
            if test != range_start - 1:
                break
            range_start -= 1
        index_start -= 1

    # Merge to the right side.
    while index_stop < ranges_len:
        test = ranges[index_stop]
        if isinstance(test, tuple):
            if test[0] != range_stop + 1:
                break
            range_stop = test[1]
        else:
            if test != range_stop + 1:
                break
            range_stop += 1
        index_stop += 1

    if range_stop - range_start < 2:
        ranges.insert(index, item)
    else:
        ranges[index_start:index_stop] = ((range_start, range_stop),)


def remove_from_ranges(ranges, item):
    """Remove item from ranges, reorganizing as needed."""

    index = find_ranges_index(ranges, item)
    ranges_len = len(ranges)
    if index < ranges_len:
        test = ranges[index]
        if isinstance(test, tuple):
            range_start, range_stop = test
            if item >= range_start:
                # Handle right side of the range (and replace original item).
                if range_stop < item + 3:
                    ranges[index:index + 1] = range(item + 1, range_stop + 1)
                else:
                    ranges[index:index + 1] = ((item + 1, range_stop),)

                # Handle left side of the range.
                if range_start > item - 3:
                    if range_start != item:
                        ranges[index:index] = range(range_start, item)
                else:
                    ranges[index:index] = ((range_start, item - 1),)
        elif item == test:
            del ranges[index]



import time
import os
import json
import logging
import re

from twisted.internet.defer import inlineCallbacks, returnValue

from landscape.accumulate import Accumulator
from landscape.lib.monitor import CoverageMonitor
from landscape.lib.twisted_util import spawn_process
from landscape.monitor.plugin import MonitorPlugin


class CephUsage(MonitorPlugin):
    """
    Plugin that captures Ceph usage information. This only works if the client
    runs on one of the Ceph monitor nodes, and it noops otherwise.

    The plugin requires the 'ceph' command to be available, which is run with a
    config file in <data_path>/ceph-client/ceph.landscape-client.conf with the
    following config:

    [global]
    auth supported = cephx
    keyring = <keyring-file>
    mon host = <ip>:6789

    The configured keyring can be generated with:

    ceph-authtool <keyring-file> --create-keyring
        --name=client.landscape-client --add-key=<key>

    The landscape-client charm automatically provides the client configuration
    and key when deployed as subordinate of a ceph node.
    """

    persist_name = "ceph-usage"
    scope = "storage"
    # Prevent the Plugin base-class from scheduling looping calls.
    run_interval = None

    _usage_regexp = re.compile(
        ".*pgmap.*data, (\d+) MB used, (\d+) MB / (\d+) MB avail.*",
        flags=re.S)

    def __init__(self, interval=30, monitor_interval=60 * 60,
                 create_time=time.time):
        self._interval = interval
        self._monitor_interval = monitor_interval
        self._ceph_usage_points = []
        self._ceph_ring_id = None
        self._create_time = create_time
        self._ceph_config = None

    def register(self, registry):
        super(CephUsage, self).register(registry)
        self._ceph_config = os.path.join(
            self.registry.config.data_path, "ceph-client",
            "ceph.landscape-client.conf")

        self._accumulate = Accumulator(self._persist, self._interval)
        self._monitor = CoverageMonitor(
            self._interval, 0.8, "Ceph usage snapshot",
            create_time=self._create_time)

        self.registry.reactor.call_every(self._interval, self.run)
        self.registry.reactor.call_every(
            self._monitor_interval, self._monitor.log)
        self.registry.reactor.call_on("stop", self._monitor.log, priority=2000)
        self.call_on_accepted("ceph-usage", self.send_message, True)

    def create_message(self):
        ceph_points = self._ceph_usage_points
        ring_id = self._ceph_ring_id
        self._ceph_usage_points = []
        return {"type": "ceph-usage", "ceph-usages": ceph_points,
                "ring-id": ring_id}

    def send_message(self, urgent=False):
        message = self.create_message()
        if message["ceph-usages"] and message["ring-id"] is not None:
            self.registry.broker.send_message(message, self._session_id,
                                              urgent=urgent)

    def exchange(self, urgent=False):
        self.registry.broker.call_if_accepted(
            "ceph-usage", self.send_message, urgent)

    @inlineCallbacks
    def run(self):
        self._monitor.ping()

        # Check if a ceph config file is available. If it's not , it's not a
        # ceph machine or ceph is set up yet. No need to run anything in this
        # case.
        if self._ceph_config is None or not os.path.exists(self._ceph_config):
            returnValue(None)

        # Extract the ceph ring Id and cache it.
        if self._ceph_ring_id is None:
            self._ceph_ring_id = yield self._get_ceph_ring_id()

        new_timestamp = int(self._create_time())
        new_ceph_usage = yield self._get_ceph_usage()

        step_data = None
        if new_ceph_usage is not None:
            step_data = self._accumulate(
                new_timestamp, new_ceph_usage, "ceph-usage-accumulator")

        if step_data is not None:
            self._ceph_usage_points.append(step_data)

    def _get_ceph_usage(self):
        """
        Grab the ceph usage data by parsing the output of the "ceph status"
        command output.
        """

        def parse(output):
            if output is None:
                return None

            result = self._usage_regexp.match(output)
            if not result:
                logging.error("Could not parse command output: '%s'." % output)
                return None

            (used, available, total) = result.groups()
            # Note: used + available is NOT equal to total (there is some used
            # space for duplication and system info etc...)
            filled = int(total) - int(available)
            return filled / float(total)

        return self._get_status_command_output().addCallback(parse)

    def _get_status_command_output(self):
        return self._run_ceph_command("status")

    def _get_ceph_ring_id(self):
        """Extract ceph ring id from ceph command output."""

        def parse(output):
            if output is None:
                return  None

            try:
                quorum_status = json.loads(output)
                ring_id = quorum_status["monmap"]["fsid"]
            except:
                logging.error(
                    "Could not get ring_id from output: '%s'." % output)
                return None
            return ring_id

        return self._get_quorum_command_output().addCallback(parse)

    def _get_quorum_command_output(self):
        return self._run_ceph_command("quorum_status")

    def _run_ceph_command(self, *args):
        """
        Run the ceph command with the specified options using landscape ceph
        key.  The keyring is expected to contain a configuration stanza with a
        key for the "client.landscape-client" id.
        """
        params = ["--conf", self._ceph_config, "--id", "landscape-client"]
        params.extend(args)
        deferred = spawn_process("ceph", args=params)
        # If the command line client isn't available, we assume it's not a ceph
        # monitor machine.
        deferred.addCallback(
            lambda (out, err, code): out if code == 0 else None)
        return deferred



import os

from landscape.lib.disk import (
    get_filesystem_for_path, get_mount_info, is_device_removable,
    _get_device_removable_file_path)
from landscape.tests.helpers import LandscapeTest


class DiskUtilitiesTest(LandscapeTest):

    def setUp(self):
        super(DiskUtilitiesTest, self).setUp()
        self.mount_file = self.makeFile("")
        self.stat_results = {}

    def statvfs(self, point):
        """
        Return the requested mount point information. If C{read_access} was
        set to C{False} when this mount point was created, then we raise an
        exception to simulate a permission denied error.
        """
        if self.read_access:
            return self.stat_results[point]
        else:
            raise OSError("Permission denied")

    def set_mount_points(self, points, read_access=True):
        """
        This method prepares a fake mounts file containing the
        mount points specified in the C{points} list of strings. This file
        can then be used by referencing C{self.mount_file}.

        If C{read_access} is set to C{False}, then all mount points will
        yield a permission denied error when inspected.
        """
        self.read_access = read_access
        content = "\n".join("/dev/sda%d %s ext4 rw 0 0" % (i, point)
                            for i, point in enumerate(points))
        f = open(self.mount_file, "w")
        f.write(content)
        f.close()
        for point in points:
            self.stat_results[point] = (4096, 0, 1000, 500, 0, 0, 0, 0, 0)

    def test_get_filesystem_for_path(self):
        self.set_mount_points(["/"])
        info = get_filesystem_for_path("/", self.mount_file, self.statvfs)
        self.assertEqual(info["mount-point"], "/")

    def test_get_filesystem_subpath(self):
        self.set_mount_points(["/"])
        self.stat_results["/"] = (4096, 0, 1000, 500, 0, 0, 0, 0, 0)
        info = get_filesystem_for_path("/home", self.mount_file, self.statvfs)
        self.assertEqual(info["mount-point"], "/")

    def test_get_filesystem_subpath_closest(self):
        self.set_mount_points(["/", "/home"])
        info = get_filesystem_for_path("/home", self.mount_file, self.statvfs)
        self.assertEqual(info["mount-point"], "/home")

    def test_get_filesystem_subpath_not_stupid(self):
        self.set_mount_points(["/", "/ho"])
        info = get_filesystem_for_path("/home", self.mount_file, self.statvfs)
        self.assertEqual(info["mount-point"], "/")

    def test_symlink_home(self):
        symlink_path = self.makeFile()
        os.symlink("/foo/bar", symlink_path)
        self.addCleanup(os.remove, symlink_path)
        self.set_mount_points(["/", "/foo"])
        info = get_filesystem_for_path(symlink_path,
                                       self.mount_file, self.statvfs)
        self.assertEqual(info["mount-point"], "/foo")

    def test_ignore_unreadable_mount_point(self):
        """
        We should ignore mountpoints which are unreadable by the user who
        is logging in.
        """
        self.set_mount_points(["/secret"], read_access=False)
        info = get_filesystem_for_path(
            "/secret", self.mount_file, self.statvfs)
        self.assertIdentical(info, None)

    def test_ignore_unmounted_and_virtual_mountpoints(self):
        """
        Make sure autofs and virtual mountpoints are ignored. This is to
        ensure non-regression on bug #1045374.
        """
        self.read_access = True
        content = "\n".join(["auto_direct /opt/whatever autofs",
                             "none /run/lock tmpfs",
                             "proc /proc proc",
                             "/dev/sda1 /home ext4"])

        f = open(self.mount_file, "w")
        f.write(content)
        f.close()

        self.stat_results["/home"] = (4096, 0, 1000, 500, 0, 0, 0, 0, 0)

        result = [x for x in get_mount_info(self.mount_file, self.statvfs)]
        expected = {"device": "/dev/sda1", "mount-point": "/home",
                    "filesystem": "ext4", "total-space": 3, "free-space": 1}
        self.assertEqual([expected], result)


class RemovableDiskTest(LandscapeTest):

    def test_wb_get_device_removable_file_path(self):
        """
        When passed a device in /dev, the L{_get_device_removable_file_path}
        function returns the corresponding removable file path in /sys/block.
        """
        device = "/dev/sdb"
        expected = "/sys/block/sdb/removable"

        is_link_mock = self.mocker.replace(os.path.islink)
        is_link_mock(device)
        self.mocker.result(False)
        self.mocker.replay()

        result = _get_device_removable_file_path(device)
        self.assertEqual(expected, result)

    def test_wb_get_device_removable_file_path_with_partition(self):
        """
        When passed a device in /dev with a partition number, the
        L{_get_device_removable_file_path} function returns the corresponding
        removable file path in /sys/block.
        """
        device = "/dev/sdb1"
        expected = "/sys/block/sdb/removable"

        is_link_mock = self.mocker.replace(os.path.islink)
        is_link_mock(device)
        self.mocker.result(False)
        self.mocker.replay()

        result = _get_device_removable_file_path(device)
        self.assertEqual(expected, result)

    def test_wb_get_device_removable_file_path_without_dev(self):
        """
        When passed a device name (not the whole path), the
        L{_get_device_removable_file_path} function returns the corresponding
        removable file path in /sys/block.
        """
        device = "sdb1"
        expected = "/sys/block/sdb/removable"

        is_link_mock = self.mocker.replace(os.path.islink)
        is_link_mock(device)
        self.mocker.result(False)
        self.mocker.replay()

        result = _get_device_removable_file_path(device)
        self.assertEqual(expected, result)

    def test_wb_get_device_removable_file_path_with_symlink(self):
        """
        When the device path passed to L{_get_device_removable_file_path} is a
        symlink (it's the case when disks are mounted by uuid or by label),
        the L{_get_device_removable_file_path} function returns the proper
        corresponding file path in /sys/block.
        """
        device = "/dev/disk/by-uuid/8b2ec410-ebd2-49ec-bb3c-b8b13effab08"

        readlink_mock = self.mocker.replace(os.readlink)
        readlink_mock(device)
        self.mocker.result("../../sda1")

        is_link_mock = self.mocker.replace(os.path.islink)
        is_link_mock(device)
        self.mocker.result(True)

        self.mocker.replay()

        expected = "/sys/block/sda/removable"
        result = _get_device_removable_file_path(device)
        self.assertEqual(expected, result)

    def test_wb_get_device_removable_file_path_raid_device(self):
        """
        When passed a more exotic device file, like for example a raid device
        (e.g. /dev/cciss/c0d1p1), the L{_get_device_removable_file_path}
        function does not fail, and returns the expected
        /sys/block/<device>/removable path.
        """
        device = "/dev/cciss/c0d0p0"
        # The expected path does not exists, but it doesn't matter here.
        expected = "/sys/block/c/removable"

        is_link_mock = self.mocker.replace(os.path.islink)
        is_link_mock(device)
        self.mocker.result(False)

        self.mocker.replay()

        result = _get_device_removable_file_path(device)
        self.assertEqual(expected, result)

    def test_is_device_removable(self):
        """
        Given the path to a file, determine if it means the device is removable
        or not.
        """
        device = "/dev/sdb1"
        path = self.makeFile("1")

        removable_mock = self.mocker.replace(_get_device_removable_file_path)
        removable_mock(device)
        self.mocker.result(path)
        self.mocker.replay()

        self.assertTrue(is_device_removable(device))

    def test_is_device_removable_false(self):
        """
        Given the path to a file, determine if it means the device is removable
        or not.
        """
        device = "/dev/sdb1"
        path = self.makeFile("0")

        removable_mock = self.mocker.replace(_get_device_removable_file_path)
        removable_mock(device)
        self.mocker.result(path)
        self.mocker.replay()

        self.assertFalse(is_device_removable(device))

    def test_is_device_removable_garbage(self):
        """
        Given the path to a file, determine if it means the device is removable
        or not.
        """
        device = "/dev/sdb1"
        path = self.makeFile("Some garbage")

        removable_mock = self.mocker.replace(_get_device_removable_file_path)
        removable_mock(device)
        self.mocker.result(path)
        self.mocker.replay()

        self.assertFalse(is_device_removable(device))

    def test_is_device_removable_path_doesnt_exist(self):
        """
        When given a non-existing path, report the device as not removable.
        """
        device = "/dev/sdb1"
        path = "/what/ever"

        removable_mock = self.mocker.replace(_get_device_removable_file_path)
        removable_mock(device)
        self.mocker.result(path)
        self.mocker.replay()

        self.assertFalse(is_device_removable(device))

    def test_is_removable_raid_device(self):
        """
        When passed the path to a raid device (e.g. /dev/cciss/c0d0p0), the
        is_device_removable function returns False.
        """
        device = "/dev/cciss/c0d1p1"

        is_link_mock = self.mocker.replace(os.path.islink)
        is_link_mock(device)
        self.mocker.result(False)
        self.mocker.replay()

        self.assertFalse(is_device_removable(device))

    def test_is_device_removable_memory_card(self):
        """
        The kernel/udev currently consider memory cards such as SD cards as non
        removable
        """
        device = "/dev/mmcblk0p1"  # Device 0, parition 1
        self.assertTrue(is_device_removable(device))



"""
Handle client registration against the server.

When the service is started for the first time it connects to the server
as a new client without providing any identification credentials, and the
server replies with the available registration mechanisms. At this point
the machinery in this module will notice that we have no identification
credentials yet and that the server accepts registration messages, so it
will craft an appropriate one and send it out.
"""
import time
import logging
import socket

from twisted.internet.defer import Deferred

from landscape.message_schemas import juju_data

from landscape.lib.bpickle import loads
from landscape.lib.log import log_failure
from landscape.lib.juju import get_juju_info
from landscape.lib.fetch import fetch, FetchError
from landscape.lib.tag import is_valid_tag_list
from landscape.lib.network import get_fqdn
from landscape.lib.vm_info import get_vm_info, get_container_info


EC2_HOST = "169.254.169.254"
EC2_API = "http://%s/latest" % (EC2_HOST,)


class InvalidCredentialsError(Exception):
    """
    Raised when an invalid account title and/or registration key
    is used with L{RegistrationManager.register}.
    """


def persist_property(name):

    def get(self):
        return self._persist.get(name)

    def set(self, value):
        self._persist.set(name, value)

    return property(get, set)


def config_property(name):

    def get(self):
        return getattr(self._config, name)

    return property(get)


class Identity(object):
    """Maintains details about the identity of this Landscape client.

    @ivar secure_id: A server-provided ID for secure message exchange.
    @ivar insecure_id: Non-secure server-provided ID, mainly used with
        the ping server.
    @ivar computer_title: See L{BrokerConfiguration}.
    @ivar account_name: See L{BrokerConfiguration}.
    @ivar registration_password: See L{BrokerConfiguration}.
    @ivar tags: See L{BrokerConfiguration}

    @param config: A L{BrokerConfiguration} object, used to set the
        C{computer_title}, C{account_name} and C{registration_password}
        instance variables.
    """

    secure_id = persist_property("secure-id")
    insecure_id = persist_property("insecure-id")
    computer_title = config_property("computer_title")
    account_name = config_property("account_name")
    registration_key = config_property("registration_key")
    tags = config_property("tags")
    access_group = config_property("access_group")

    def __init__(self, config, persist):
        self._config = config
        self._persist = persist.root_at("registration")


class RegistrationHandler(object):
    """
    An object from which registration can be requested of the server,
    and which will handle forced ID changes from the server.

    L{register} should be used to perform initial registration.
    """

    def __init__(self, config, identity, reactor, exchange, pinger,
                 message_store, fetch_async=None):
        self._config = config
        self._identity = identity
        self._reactor = reactor
        self._exchange = exchange
        self._pinger = pinger
        self._message_store = message_store
        self._reactor.call_on("run", self._fetch_ec2_data)
        self._reactor.call_on("run", self._get_juju_data)
        self._reactor.call_on("pre-exchange", self._handle_pre_exchange)
        self._reactor.call_on("exchange-done", self._handle_exchange_done)
        self._exchange.register_message("set-id", self._handle_set_id)
        self._exchange.register_message("unknown-id", self._handle_unknown_id)
        self._exchange.register_message("registration",
                                        self._handle_registration)
        self._should_register = None
        self._fetch_async = fetch_async
        self._otp = None
        self._ec2_data = None
        self._juju_data = None

    def should_register(self):
        id = self._identity
        if id.secure_id:
            # We already have a secure ID, no need to register
            return False
        if self._config.cloud:
            return self._message_store.accepts("register-cloud-vm")
        elif self._config.provisioning_otp:
            return self._message_store.accepts("register-provisioned-machine")
        return bool(id.computer_title and id.account_name
                    and self._message_store.accepts("register"))

    def register(self):
        """
        Attempt to register with the Landscape server.

        @return: A L{Deferred} which will either be fired with None if
            registration was successful or will fail with an
            L{InvalidCredentialsError} if not.
        """
        self._identity.secure_id = None
        self._identity.insecure_id = None
        result = RegistrationResponse(self._reactor).deferred
        self._exchange.exchange()
        return result

    def _get_juju_data(self):
        """Load Juju information."""
        juju_info = get_juju_info(self._config)
        if juju_info is None:
            return None
        self._juju_data = dict(
            (key, juju_info[key]) for key in juju_data)

    def _get_data(self, path, accumulate):
        """
        Get data at C{path} on the EC2 API endpoint, and add the result to the
        C{accumulate} list.
        """
        return self._fetch_async(EC2_API + path).addCallback(accumulate.append)

    def _fetch_ec2_data(self):
        """Retrieve available EC2 information, if in a EC2 compatible cloud."""
        id = self._identity
        if self._config.cloud and not id.secure_id:
            # Fetch data from the EC2 API, to be used later in the registration
            # process
            # We ignore errors from user-data because it's common for the
            # URL to return a 404 when the data is unavailable.
            ec2_data = []
            deferred = self._fetch_async(EC2_API + "/user-data").addErrback(
                log_failure).addCallback(ec2_data.append)
            paths = [
                "/meta-data/instance-id",
                "/meta-data/reservation-id",
                "/meta-data/local-hostname",
                "/meta-data/public-hostname",
                "/meta-data/ami-launch-index",
                "/meta-data/ami-id",
                "/meta-data/local-ipv4",
                "/meta-data/public-ipv4"]
            # We're not using a DeferredList here because we want to keep the
            # number of connections to the backend minimal. See lp:567515.
            for path in paths:
                deferred.addCallback(
                    lambda ignore, path=path: self._get_data(path, ec2_data))
            # Special case the ramdisk retrieval, because it may not be present
            deferred.addCallback(
                lambda ignore: self._fetch_async(
                    EC2_API + "/meta-data/ramdisk-id").addErrback(log_failure))
            deferred.addCallback(ec2_data.append)
            # And same for kernel
            deferred.addCallback(
                lambda ignore: self._fetch_async(
                    EC2_API + "/meta-data/kernel-id").addErrback(log_failure))
            deferred.addCallback(ec2_data.append)

            def record_data(ignore):
                """Record the instance data returned by the EC2 API."""
                (raw_user_data, instance_key, reservation_key,
                 local_hostname, public_hostname, launch_index,
                 ami_key, local_ip, public_ip, ramdisk_key,
                 kernel_key) = ec2_data
                self._ec2_data = {
                    "instance_key": instance_key,
                    "reservation_key": reservation_key,
                    "local_hostname": local_hostname,
                    "public_hostname": public_hostname,
                    "launch_index": launch_index,
                    "kernel_key": kernel_key,
                    "ramdisk_key": ramdisk_key,
                    "image_key": ami_key,
                    "public_ipv4": public_ip,
                    "local_ipv4": local_ip}
                for k, v in self._ec2_data.items():
                    if v is None and k in ("ramdisk_key", "kernel_key"):
                        continue
                    self._ec2_data[k] = v.decode("utf-8")
                self._ec2_data["launch_index"] = int(
                    self._ec2_data["launch_index"])

                if self._config.otp:
                    self._otp = self._config.otp
                    return
                instance_data = _extract_ec2_instance_data(
                    raw_user_data, int(launch_index))
                if instance_data is not None:
                    self._otp = instance_data["otp"]
                    exchange_url = instance_data["exchange-url"]
                    ping_url = instance_data["ping-url"]
                    self._exchange._transport.set_url(exchange_url)
                    self._config.url = exchange_url
                    self._config.ping_url = ping_url
                    if "ssl-ca-certificate" in instance_data:
                        from landscape.configuration import \
                            store_public_key_data
                        public_key_file = store_public_key_data(
                            self._config, instance_data["ssl-ca-certificate"])
                        self._config.ssl_public_key = public_key_file
                        self._exchange._transport._pubkey = public_key_file
                    self._config.write()

            def log_error(error):
                log_failure(error, msg="Got error while fetching meta-data: %r"
                            % (error.value,))

            deferred.addCallback(record_data)
            deferred.addErrback(log_error)

    def _handle_exchange_done(self):
        """Registered handler for the C{"exchange-done"} event.

        If we are not registered yet, schedule another message exchange.

        The first exchange made us accept the message type "register", so
        the next "pre-exchange" event will make L{_handle_pre_exchange}
        queue a registration message for delivery.
        """
        if self.should_register() and not self._should_register:
            self._exchange.exchange()

    def _handle_pre_exchange(self):
        """
        An exchange is about to happen.  If we don't have a secure id already
        set, and we have the needed information available, queue a registration
        message with the server.

        A computer can fall into several categories:
            - a "cloud VM"
            - a "normal" computer
            - a "provisionned machine".

        Furthermore, Cloud VMs can be registered with either a One Time
        Password (OTP), or with a normal registration password.
        """
        registration_failed = False

        # The point of storing this flag is that if we should *not* register
        # now, and then after the exchange we *should*, we schedule an urgent
        # exchange again.  Without this flag we would just spin trying to
        # connect to the server when something is clearly preventing the
        # registration.
        self._should_register = self.should_register()
        if not self._should_register:
            return

        # These are just to shorten the code.
        identity = self._identity
        account_name = identity.account_name
        tags = identity.tags
        group = identity.access_group
        registration_key = identity.registration_key

        self._message_store.delete_all_messages()

        if not is_valid_tag_list(tags):
            tags = None
            logging.error("Invalid tags provided for cloud registration.")

        message = {"type": None,  # either "register" or "register-cloud-vm"
                   "otp": None,
                   "hostname": get_fqdn(),
                   "account_name": identity.account_name,
                   "registration_password": None,
                   "tags": tags,
                   "vm-info": get_vm_info()}

        if group:
            message["access_group"] = group

        if self._config.cloud and self._ec2_data is not None:
            # This is the "cloud VM" case.
            message["type"] = "register-cloud-vm"

            message.update(self._ec2_data)

            if self._otp:
                logging.info("Queueing message to register with OTP")
                message["otp"] = self._otp

            elif account_name:
                with_tags = "and tags %s " % tags if tags else ""
                with_group = "in access group '%s' " % group if group else ""
                logging.info(
                    u"Queueing message to register with account %r %s%s"
                    u"as an EC2 instance." % (
                        account_name, with_group, with_tags))
                message["registration_password"] = registration_key

            else:
                registration_failed = True

        elif account_name:
            # The computer is a normal computer, possibly a container.
            with_word = "with" if bool(registration_key) else "without"
            with_tags = "and tags %s " % tags if tags else ""
            with_group = "in access group '%s' " % group if group else ""

            logging.info(u"Queueing message to register with account %r %s%s"
                         "%s a password." % (account_name, with_group,
                                             with_tags, with_word))
            message["type"] = "register"
            message["computer_title"] = identity.computer_title
            message["registration_password"] = identity.registration_key
            message["container-info"] = get_container_info()

            if self._juju_data is not None:
                message["juju-info"] = self._juju_data

        elif self._config.provisioning_otp:
            # This is a newly provisionned machine.
            # In this case message is overwritten because it's much simpler
            logging.info(u"Queueing message to register with OTP as a"
                         u" newly provisioned machine.")
            message = {"type": "register-provisioned-machine",
                       "otp": self._config.provisioning_otp}
        else:
            registration_failed = True

        if registration_failed:
            self._reactor.fire("registration-failed")
        else:
            self._exchange.send(message)

    def _handle_set_id(self, message):
        """Registered handler for the C{"set-id"} event.

        Record and start using the secure and insecure IDs from the given
        message.

        Fire C{"registration-done"} and C{"resynchronize-clients"}.
        """
        id = self._identity
        id.secure_id = message.get("id")
        id.insecure_id = message.get("insecure-id")
        logging.info("Using new secure-id ending with %s for account %s.",
                     id.secure_id[-10:], id.account_name)
        logging.debug("Using new secure-id: %s", id.secure_id)
        self._reactor.fire("registration-done")
        self._reactor.fire("resynchronize-clients")

    def _handle_registration(self, message):
        if message["info"] == "unknown-account":
            self._reactor.fire("registration-failed")

    def _handle_unknown_id(self, message):
        id = self._identity
        clone = message.get("clone-of")
        if clone is None:
            logging.info("Client has unknown secure-id for account %s."
                         % id.account_name)
        else:
            logging.info("Client is clone of computer %s" % clone)
            # Set a new computer title so when a registration request will be
            # made, the pending computer UI will indicate that this is a clone
            # of another computer. There's no need to persist the changes since
            # a new registration will be requested immediately.
            if clone == self._config.computer_title:
                title = "%s (clone)" % self._config.computer_title
            else:
                title = "%s (clone of %s)" % (self._config.computer_title,
                                              clone)
            self._config.computer_title = title
        id.secure_id = None
        id.insecure_id = None


class RegistrationResponse(object):
    """A helper for dealing with the response of a single registration request.

    @ivar deferred: The L{Deferred} that will be fired as per
        L{RegistrationHandler.register}.
    """

    def __init__(self, reactor):
        self._reactor = reactor
        self._done_id = reactor.call_on("registration-done", self._done)
        self._failed_id = reactor.call_on("registration-failed", self._failed)
        self.deferred = Deferred()

    def _cancel_calls(self):
        self._reactor.cancel_call(self._done_id)
        self._reactor.cancel_call(self._failed_id)

    def _done(self):
        self.deferred.callback(None)
        self._cancel_calls()

    def _failed(self):
        self.deferred.errback(InvalidCredentialsError())
        self._cancel_calls()


def _extract_ec2_instance_data(raw_user_data, launch_index):
    """
    Given the raw string of EC2 User Data, parse it and return the dict of
    instance data for this particular instance.

    If the data can't be parsed, a debug message will be logged and None
    will be returned.
    """
    try:
        user_data = loads(raw_user_data)
    except ValueError:
        logging.debug("Got invalid user-data %r" % (raw_user_data,))
        return

    if not isinstance(user_data, dict):
        logging.debug("user-data %r is not a dict" % (user_data,))
        return
    for key in "otps", "exchange-url", "ping-url":
        if key not in user_data:
            logging.debug("user-data %r doesn't have key %r."
                          % (user_data, key))
            return
    if len(user_data["otps"]) <= launch_index:
        logging.debug("user-data %r doesn't have OTP for launch index %d"
                      % (user_data, launch_index))
        return
    instance_data = {"otp": user_data["otps"][launch_index],
                     "exchange-url": user_data["exchange-url"],
                     "ping-url": user_data["ping-url"]}
    if "ssl-ca-certificate" in user_data:
        instance_data["ssl-ca-certificate"] = user_data["ssl-ca-certificate"]
    return instance_data


def _wait_for_network():
    """
    Keep trying to connect to the EC2 metadata server until it becomes
    accessible or until five minutes pass.

    This is necessary because the networking init script on Ubuntu is
    asynchronous; the network may not actually be up by the time the
    landscape-client init script is invoked.
    """
    timeout = 5 * 60
    port = 80

    start = time.time()
    while True:
        s = socket.socket()
        try:
            s.connect((EC2_HOST, port))
            s.close()
            return
        except socket.error:
            time.sleep(1)
            if time.time() - start > timeout:
                break


def is_cloud_managed(fetch=fetch):
    """
    Return C{True} if the machine has been started by Landscape, i.e. if we can
    find the expected data inside the EC2 user-data field.
    """
    _wait_for_network()
    try:
        raw_user_data = fetch(EC2_API + "/user-data",
                              connect_timeout=5)
        launch_index = fetch(EC2_API + "/meta-data/ami-launch-index",
                             connect_timeout=5)
    except FetchError:
        return False
    instance_data = _extract_ec2_instance_data(
        raw_user_data, int(launch_index))
    return instance_data is not None



import sys

from landscape.textmessage import (
    AcceptedTypeError, EmptyMessageError, got_accepted_types, get_message,
    send_message)
from landscape.tests.helpers import (
    LandscapeTest, FakeBrokerServiceHelper, StandardIOHelper)


class SendMessageTest(LandscapeTest):

    helpers = [StandardIOHelper, FakeBrokerServiceHelper]

    def test_send_message(self):
        """
        L{send_message} should send a message of type
        C{text-message} to the landscape dbus messaging service.
        """
        service = self.broker_service
        service.message_store.set_accepted_types(["text-message"])

        result = send_message(u"Hi there!", self.remote)

        def got_result(result):
            messages = service.message_store.get_pending_messages()
            self.assertEqual(len(messages), 1)
            self.assertMessage(messages[0], {"type": "text-message",
                                             "message": u"Hi there!"})
            self.assertTrue(service.exchanger.is_urgent())
        return result.addCallback(got_result)

    def test_got_accepted_types_without_text_message_type(self):
        """
        If 'text-message' isn't in the list of accepted types an
        L{AcceptedTypeError} is raised.
        """
        self.assertRaises(AcceptedTypeError, got_accepted_types, (),
                          self.remote, ())

    def test_got_accepted_types(self):
        """
        If 'text-message' is an accepted type a message should be
        retrieved from the user and sent to the broker.
        """
        service = self.broker_service
        service.message_store.set_accepted_types(["text-message"])

        input = u"Foobl\N{HIRAGANA LETTER A}"
        self.stdin.write(input.encode("UTF-8"))
        self.stdin.seek(0, 0)

        def got_result(result):
            messages = service.message_store.get_pending_messages()
            self.assertEqual(len(messages), 1)
            self.assertMessage(messages[0],
                               {"type": "text-message",
                                "message": u"Foobl\N{HIRAGANA LETTER A}"})

        d = got_accepted_types(["text-message"], self.remote, ())
        d.addCallback(got_result)
        return d


class ScriptTest(LandscapeTest):

    helpers = [StandardIOHelper]

    def test_get_message(self):
        """
        A message should be properly decoded from the command line arguments.
        """
        message = get_message(
            ["landscape-message",
             u"\N{HIRAGANA LETTER A}".encode(sys.stdin.encoding), "a!"])
        self.assertEqual(message, u"\N{HIRAGANA LETTER A} a!")

    def test_get_message_stdin(self):
        """
        If no arguments are specified then the message should be read
        from stdin.
        """
        input = u"Foobl\N{HIRAGANA LETTER A}"
        self.stdin.write(input.encode("UTF-8"))
        self.stdin.seek(0, 0)
        message = get_message(["landscape-message"])
        self.assertEqual(self.stdout.getvalue(),
                         "Please enter your message, and send EOF "
                         "(Control + D after newline) when done.\n")
        self.assertEqual(message, input)

    def test_get_empty_message_stdin(self):
        """
        If no arguments are specified then the message should be read
        from stdin.
        """
        self.assertRaises(EmptyMessageError,
                          get_message, ["landscape-message"])

    def test_get_message_without_encoding(self):
        """
        If sys.stdin.encoding is None, it's likely a pipe, so try to
        decode it as UTF-8 by default.
        """
        encoding = sys.stdin.encoding
        sys.stdin.encoding = None
        try:
            message = get_message(
                ["landscape-message",
                 u"\N{HIRAGANA LETTER A}".encode("UTF-8"), "a!"])
        finally:
            sys.stdin.encoding = encoding
        self.assertEqual(message, u"\N{HIRAGANA LETTER A} a!")



import inspect
import logging


def format_object(object):
    """
    Returns a fully-qualified name for the specified object, such as
    'landscape.log.format_object()'.
    """
    if inspect.ismethod(object):
        # FIXME If the method is implemented on a base class of
        # object's class, the module name and function name will be
        # from the base class and the method's class name will be from
        # object's class.
        name = repr(object).split(" ")[2]
        return "%s.%s()" % (object.__module__, name)
    elif inspect.isfunction(object):
        name = repr(object).split(" ")[1]
        return "%s.%s()" % (object.__module__, name)
    return "%s.%s" % (object.__class__.__module__, object.__class__.__name__)


def format_delta(seconds):
    if not seconds:
        seconds = 0.0
    return "%.02fs" % float(seconds)


def format_percent(percent):
    if not percent:
        percent = 0.0
    return "%.02f%%" % float(percent)

def rotate_logs():
    """
    This closes and reopens the underlying files in the logging module's
    root logger. If called after logrotate (or something similar) has
    moved the old log file out of the way, this will start writing to a new
    new log file...
    """
    for handler in logging.getLogger().handlers:
        if isinstance(handler, logging.FileHandler):
            handler.acquire()
            try:
                handler.stream.close()
                handler.stream = open(handler.baseFilename,
                                      handler.mode)
            finally:
                handler.release()
    logging.info("Landscape Logs rotated")



import logging
import signal

from twisted.application.service import Application, Service
from twisted.application.app import startApplication

from landscape.log import rotate_logs
from landscape.reactor import LandscapeReactor
from landscape.deployment import get_versioned_persist, init_logging


class LandscapeService(Service, object):
    """Utility superclass for defining Landscape services.

    This sets up the reactor and L{Persist} object.

    @cvar service_name: The lower-case name of the service. This is used to
        generate the bpickle and the Unix socket filenames.
    @ivar config: A L{Configuration} object.
    @ivar reactor: A L{LandscapeReactor} object.
    @ivar persist: A L{Persist} object, if C{persist_filename} is defined.
    @ivar factory: A L{LandscapeComponentProtocolFactory}, it must be provided
        by instances of sub-classes.
    """
    reactor_factory = LandscapeReactor
    persist_filename = None

    def __init__(self, config):
        self.config = config
        try:
            from landscape.lib import bpickle_dbus
        except ImportError:
            pass
        else:
            bpickle_dbus.install()
        self.reactor = self.reactor_factory()
        if self.persist_filename:
            self.persist = get_versioned_persist(self)
        if not (self.config is not None and self.config.ignore_sigusr1):
            from twisted.internet import reactor
            signal.signal(
                signal.SIGUSR1,
                lambda signal, frame: reactor.callFromThread(rotate_logs))

    def startService(self):
        Service.startService(self)
        logging.info("%s started with config %s" % (
            self.service_name.capitalize(), self.config.get_config_filename()))

    def stopService(self):
        # We don't need to call port.stopListening(), because the reactor
        # shutdown sequence will do that for us.
        Service.stopService(self)
        logging.info("%s stopped with config %s" % (
            self.service_name.capitalize(), self.config.get_config_filename()))


def run_landscape_service(configuration_class, service_class, args):
    """Run a Landscape service.

    This function instantiates the specified L{LandscapeService} subclass and
    attaches the resulting service object to a Twisted C{Application}.  After
    that it starts the Twisted L{Application} and calls the
    L{LandscapeReactor.run} method of the L{LandscapeService}'s reactor.

    @param configuration_class: The service-specific subclass of
        L{Configuration} used to parse C{args} and build the C{service_class}
        object.
    @param service_class: The L{LandscapeService} subclass to create and start.
    @param args: Command line arguments used to initialize the configuration.
    """
    # Let's consider adding this:
    # from twisted.python.log import (
    #     startLoggingWithObserver, PythonLoggingObserver)
    # startLoggingWithObserver(PythonLoggingObserver().emit, setStdout=False)

    configuration = configuration_class()
    configuration.load(args)
    init_logging(configuration, service_class.service_name)
    application = Application("landscape-%s" % (service_class.service_name,))
    service = service_class(configuration)
    service.setServiceParent(application)

    if configuration.clones > 0:

        # Increase the timeout of AMP's MethodCalls
        # XXX: we should find a better way to expose this knot, and
        # not set it globally on the class
        from landscape.lib.amp import MethodCallSender
        MethodCallSender.timeout = 300

        # Create clones here because LandscapeReactor.__init__ would otherwise
        # cancel all scheduled delayed calls
        clones = []
        for i in range(configuration.clones):
            clone_config = configuration.clone()
            clone_config.computer_title += " Clone %d" % i
            clone_config.master_data_path = configuration.data_path
            clone_config.data_path += "-clone-%d" % i
            clone_config.log_dir += "-clone-%d" % i
            clone_config.is_clone = True
            clones.append(service_class(clone_config))

        configuration.is_clone = False

        def start_clones():
            # Spawn instances over the given time window
            start_clones_over = float(configuration.start_clones_over)
            delay = start_clones_over / configuration.clones

            for i, clone in enumerate(clones):

                def start(clone):
                    clone.setServiceParent(application)
                    clone.reactor.fire("run")

                service.reactor.call_later(delay * (i + 1), start, clone=clone)

        service.reactor.call_when_running(start_clones)

    startApplication(application, False)
    if configuration.ignore_sigint:
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    service.reactor.run()



from twisted.internet.defer import succeed

from landscape.lib.sysstats import MemoryStats


class Memory(object):

    def __init__(self, filename="/proc/meminfo"):
        self._filename = filename

    def register(self, sysinfo):
        self._sysinfo = sysinfo

    def run(self):
        memstats = MemoryStats(self._filename)
        self._sysinfo.add_header("Memory usage", "%d%%" %
                                 memstats.used_memory_percentage)
        self._sysinfo.add_header("Swap usage", "%d%%" %
                                 memstats.used_swap_percentage)
        return succeed(None)
